from django.conf import settings
import defaults


for setting in dir(defaults):
    if setting == setting.upper() and not hasattr(settings, setting):
        setattr(settings, setting, getattr(defaults, setting))



from __future__ import with_statement
import hashlib
import os
import random
import re
from os.path import isfile, isdir, getmtime, dirname, getsize, normpath, join as pjoin
from PIL import Image
from django.utils._os import safe_join
from django.utils.encoding import force_unicode, smart_str, iri_to_uri
from django.utils.safestring import mark_safe
from convert import helpers
from convert.conf import settings


re_remote = re.compile(r'^(https?|ftp):\/\/')
re_whitespace = re.compile(r'\s{2,}')
re_ext = re.compile(r'\.([a-zA-Z]{2,4})$')


class MediaFile(object):
    """
    A media file wrapper, targeted for image manipulation using
    ImageMagick convert.
    """
    def __init__(self, name):
        name = force_unicode(name)
        # remote files need to be cached first
        if re_remote.match(name):
            m = re_ext.search(name)
            ext = m and m.group(1).lower() or 'jpg' # brutally resort to jpg
            cached = get_mediafile(name, ext)
            if not cached.exists():
                try:
                    helpers.execute(settings.CONVERT_WGET_PATH,
                        ['-T', '%s' % settings.CONVERT_REMOTE_TIMEOUT,
                        '-O', cached.path, name])
                except helpers.ExecuteException:
                    # we treat this as source missing
                    pass
                else:
                    cached.set_xmp({'source': name})
            name = cached.name
        self.name = name
        path = normpath(safe_join(settings.MEDIA_ROOT, self.name))
        self.path = smart_str(path)
        path_dir = dirname(self.path)
        if not isdir(path_dir):
            os.makedirs(path_dir)

    def __unicode__(self):
        return self.name

    def _get_dimensions(self):
        if not hasattr(self, '_dimensions'):
            self._dimensions = Image.open(self.path).size
        return self._dimensions
    dimensions = property(_get_dimensions)
    
    width = property(lambda self: self.dimensions[0])
    height = property(lambda self: self.dimensions[1])

    def size(self):
        return getsize(self.path)

    def _get_tag(self):
        return settings.CONVERT_TAG % {
            'width': self.width,
            'height': self.height,
            'url': self.url,
        } 
    tag = property(_get_tag)

    def _get_relative_url(self):
        return iri_to_uri(self.name.replace('\\', '/'))
    relative_url = property(_get_relative_url)
    
    def _get_url(self):
        if not hasattr(self, '_url'):
            random.seed(self.name)
            base_url = random.choice(settings.STATIC_URLS)
            random.seed() # reset the seed
            self._url = "%s%s" % (base_url, self.relative_url)
        return self._url
    url = property(_get_url)

    def set_xmp(self, params, clear=True):
        if settings.CONVERT_SET_XMP:
            # some files cannot have XMP metadata written by exiv2, notably GIF
            try:
                helpers.write_xmp(self.path, params, clear=clear)
            except helpers.ExecuteException:
                if settings.CONVERT_EXIV2_DEBUG:
                    raise

    def write(self, data):
        with open(self.path, 'wb') as fp:
            fp.write(data)

    def exists(self):
        return isfile(self.path)

    def thumbnail(self, options):
        """
        Simple alternative syntax for thumbnails
        """
        options, ext = helpers.thumbnail_to_convert(options)
        return self.convert(options, ext=ext)

    def convert(self, options, ext=None):
        # we sanitize relation since this is base for the seed and the output
        # filename
        options = re_whitespace.sub(' ', options.strip())
        dest = get_mediafile(options + self.name, ext)
        if not dest.exists() or getmtime(dest.path) < getmtime(self.path):
            args = "%s %s %s" % (self.path, options, dest.path)
            helpers.execute(settings.CONVERT_PATH, args)
            dest.set_xmp({'source': self.name, 'relation': options})
        return dest


class EmptyMediaFile(object):
    """
    Empty instance to be used when CONVERT_DEBUG is False
    """
    url = settings.CONVERT_404_URL
    tag = settings.CONVERT_EMPTY_TAG

    def __unicode__(self):
        return ''

    def __getattr__(self, name):
        return ''


def get_mediafile(seed, ext=None):
    """
    Function to get MediaFile instance based on seed and ext
    """
    hash_ = hashlib.md5(smart_str(seed)).hexdigest()
    ext = ext or settings.CONVERT_EXTENSION
    sub_tree = os.sep.join([hash_[n:n + 2] for n in xrange(0,
            settings.CONVERT_CACHE_SUBDIR_LEVELS * 2, 2)])
    name = pjoin(settings.CONVERT_CACHE, sub_tree, "%s.%s" % (hash_, ext))
    return MediaFile(name)

def convert_solo(options, ext=None):
    options = re_whitespace.sub(' ', options.strip())
    dest = get_mediafile(options, ext)
    if not dest.exists():
        args = "%s %s" % (options, dest.path)
        helpers.execute(settings.CONVERT_PATH, args)
        dest.set_xmp({'relation': options})
    return dest



from os.path import join as pjoin
from django.conf import settings


# default output extension, you can override this on a per tag basis.
# ImageMagick convert automagically knows the format from extension
CONVERT_EXTENSION = 'jpg'

# default thumbnail quality, 0-100
CONVERT_THUMBNAIL_QUALITY = 85

# formatting of tag, xhtml anyone?
CONVERT_TAG = '<img width="%(width)s" height="%(height)s" src="%(url)s">'

# spreading media over a few domains can have positive effects on performance
STATIC_URLS = (settings.MEDIA_URL,)

# cache path relative MEDIA_ROOT
CONVERT_CACHE = 'cache'

# number of levels of subdirs to the cache dir, each level will hold 256 dirs.
# Note that if you change this all existing cache will need to be moved or
# re-cached. Comments regardning some values:
# 0: all cached files are storded right under CONVERT_CACHE
# 1: cached files will use 256 sub-directories under CONVERT_CACHE
# 2: CONVERT_CACHE will hold 256 sub-directories each with 256 sub-directories
# 3: will most likely lead to that every cached file will have its own dir
CONVERT_CACHE_SUBDIR_LEVELS = 1

# write XMP data to file, this can be useful for finding stale cache
CONVERT_SET_XMP = True

# will be returned on errors rather than an img tag with an empty source which
# is known to cause some serious problems as in requesting the paga aigain
CONVERT_EMPTY_TAG = '<img>'

# url attribute when something goes wrong, use something that returns a 404
CONVERT_404_URL = '/404/'

# path to shell commands, full path is only needed if the command is not in
# users path
CONVERT_PATH = 'convert'
CONVERT_EXIV2_PATH = 'exiv2'
CONVERT_WGET_PATH = 'wget'

# when True ConvertNode.render can raise errors
CONVERT_DEBUG = False

# when True failing to write xmp data can raise errors
CONVERT_EXIV2_DEBUG = False

# timeout in seconds for caching remote files locally, used by wget
CONVERT_REMOTE_TIMEOUT = 5



from django.template import Library, Node, TemplateSyntaxError
from django.utils.encoding import force_unicode
from convert.base import MediaFile, EmptyMediaFile, convert_solo
from convert.conf import settings


register = Library()


class ConvertBaseNode(Node):
    def error(self, context):
        if settings.CONVERT_DEBUG:
            raise
        elif self.as_var:
            context[self.as_var] = EmptyMediaFile()
            return ''
        return EmptyMediaFile().tag

    def success(self, context, dest):
        if self.as_var:
            context[self.as_var] = dest
            return ''
        return dest.tag


class ThumbnailNode(ConvertBaseNode):
    def __init__(self, input_file, options, as_var):
        self.input_file = input_file
        self.options = options
        self.as_var = as_var
    
    def render(self, context):
        try:
            input_file = force_unicode(self.input_file.resolve(context))
            options = self.options.resolve(context)
            source = MediaFile(input_file)
            dest = source.thumbnail(options)
        except:
            return self.error(context)
        return self.success(context, dest)


class ConvertNode(ConvertBaseNode):
    def __init__(self, input_file, options, ext,
            as_var):
        self.input_file = input_file
        self.options = options
        self.ext = ext
        self.as_var = as_var

    def render(self, context):
        try:
            input_file = force_unicode(self.input_file.resolve(context))
            options = self.options.resolve(context)
            ext = self.ext and self.ext.resolve(context)
            if not input_file:
                dest = convert_solo(options, ext)
            else:
                source = MediaFile(input_file)
                dest = source.convert(options, ext)
        except:
            return self.error(context)
        return self.success(context, dest)


@register.tag
def thumbnail(parser, token):
    args = token.split_contents()
    invalid_syntax = TemplateSyntaxError('Invalid syntax.\nGot: %s\n'
            'Expected: thumbnail "input-file" "options" [as var]'
            % " ".join(args))
    as_var = None
    if len(args) not in (3, 5):
        raise invalid_syntax
    if args[-2] == 'as':
        as_var = args[-1]
        args = args[:-2]
    if len(args) != 3:
        raise invalid_syntax
    input_file, options = map(parser.compile_filter, args[1:])
    return ThumbnailNode(input_file, options, as_var)

@register.tag
def convert(parser, token):
    args = token.split_contents()
    invalid_syntax = TemplateSyntaxError('Invalid syntax.\nGot: %s.\n'
            'Expected: convert "input-file" "options" ["extension"] '
            '[as var]' % " ".join(args))
    as_var = None
    ext = None
    if len(args) < 3:
        raise invalid_syntax
    if args[-2] == 'as':
        as_var = args[-1]
        args = args[:-2]
    if len(args) == 4:
        ext = parser.compile_filter(args.pop(3))
    if len(args) != 3:
        raise invalid_syntax
    input_file, options = map(parser.compile_filter,
            args[1:])
    return ConvertNode(input_file, options, ext, as_var)



from os.path import abspath, dirname, join as pjoin
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


fn = abspath(pjoin(dirname(__file__), 'README.rst')) 
fp = open(fn, 'r')
long_description = fp.read()
fp.close()

setup(
    name='aino-convert',
    version='0.1.0.1',
    url='http://bitbucket.org/aino/aino-convert/',
    license='BSD',
    author='Mikko Hellsing',
    author_email='mikko@aino.se',
    description='Magick for Django',
    long_description=long_description,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Multimedia :: Graphics',
        'Framework :: Django',
    ],
    packages=[
        'convert',
        'convert.conf',
        'convert.templatetags',
    ],
    platforms='any'
)



import re
from subprocess import Popen, PIPE
from convert.conf import settings
from django.utils.encoding import smart_str


re_geometry = re.compile('^(\d+)[\^><]*x(\d+)[\^><]*$')
re_width = re.compile('^(\d+)[\^><]*$')
re_height = re.compile('^x(\d+)[\^><]*$')


class ThumbnailException(Exception):
    pass

class ExecuteException(Exception):
    pass

def execute(cmd, args):
    """
    Just a simple wrapper for executing commands. args is expected to be a
    sequence or a basestring, a basestring will be executed with shell=True.
    """
    if isinstance(args, basestring):
        args = '%s %s' % (cmd, args)
        args = smart_str(args)
        shell = True
    else:
        args.insert(0, cmd)
        args = map(smart_str, args)
        shell = False
    p = Popen(args, shell=shell, stderr=PIPE)
    retcode = p.wait()
    if retcode != 0:
        raise ExecuteException(p.stderr.read().strip())

def write_xmp(path, params, clear=True, namespace='dc'):
    """
    Used to write Adobe XMP (metadata) to file.
    """
    for k, v in params.items():
        if clear:
            execute(settings.CONVERT_EXIV2_PATH,
                    ['-M', 'del Xmp.%s.%s' % (namespace, k), path])
        if v:
            execute(settings.CONVERT_EXIV2_PATH,
                    ['-M', 'set Xmp.%s.%s %s' % (namespace, k, v), path])

def thumbnail_to_convert(value):
    """
    Helper function to generate proper convert args syntax from a simpler
    syntax. returns convert output-options and output-extension
    """
    if not len(value):
        raise ThumbnailException('Missing argument.')
    input_args = value.split(',')
    geometry = input_args.pop(0)
    # < and > mess up highlighting in html editors
    geometry = geometry.replace('gt', '>').replace('lt', '<')
    kwargs = {}
    for arg in input_args:
        arg = arg.strip().split('=')
        kwargs[arg[0]] = arg[1] if len(arg) > 1 else None
        
    def get_dimensions():
        m = re_geometry.match(geometry)
        if m:
            return m.groups()
        m = re_width.match(geometry)
        if m:
            return m.group(1), None
        m = re_height.match(geometry)
        if m:
            return None, m.group(1)
        return None

    dimensions = get_dimensions()
    if dimensions is None:
        raise ThumbnailException('Invalid geometry.')

    args = ["-strip -thumbnail '%s'" % geometry]
    if 'crop' in kwargs:
        if None in dimensions:
            raise ThumbnailException('crop needs both width and height.')
        args.insert(0, '-gravity center')
        args.append('-crop %sx%s+0+0' % dimensions)
    if kwargs.get('colorspace'):
        args.append('-colorspace %s' % kwargs['colorspace'])
    if kwargs.get('sharpen'):
        args.append('-sharpen %s' % kwargs['sharpen'])
    quality = kwargs.get('quality', settings.CONVERT_THUMBNAIL_QUALITY)
    args.append('-quality %s' % quality)
    return " ".join(args), kwargs.get('ext')
